@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.51 2002/11/08 01:10:09 hpa Exp $ */
+/* tftp-hpa: $Id: tftpd.c,v 1.57 2004/06/13 21:11:24 hpa Exp $ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -45,7 +45,7 @@ static const char *copyright UNUSED =
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
 static const char *rcsid UNUSED =
-"tftp-hpa $Id: tftpd.c,v 1.51 2002/11/08 01:10:09 hpa Exp $";
+"tftp-hpa $Id: tftpd.c,v 1.57 2004/06/13 21:11:24 hpa Exp $";
 #endif /* not lint */
 
 /*
@@ -92,8 +92,10 @@ int           timeout_quit = 0;
 sigjmp_buf    timeoutbuf;
 
 #define	PKTSIZE	MAX_SEGSIZE+4
-char	buf[PKTSIZE];
-char	ackbuf[PKTSIZE];
+char		buf[PKTSIZE];
+char		ackbuf[PKTSIZE];
+unsigned int	max_blksize = MAX_SEGSIZE;
+
 struct	sockaddr_in from;
 int	fromlen;
 off_t	tsize;
@@ -205,25 +207,37 @@ set_socket_nonblock(int fd, int flag)
 }
 
 /*
- * Receive packet with synchronous timeout
+ * Receive packet with synchronous timeout; timeout is adjusted
+ * to account for time spent waiting.
  */
 static int recv_time(int s, void *rbuf, int len, unsigned int flags,
-		     unsigned long timeout_us)
+		     unsigned long *timeout_us_p)
 {
   fd_set fdset;
-  struct timeval tmv;
+  struct timeval tmv, t0, t1;
   int rv, err;
+  unsigned long timeout_us = *timeout_us_p;
+  unsigned long timeout_left, dt;
+
+  gettimeofday(&t0, NULL);
+  timeout_left = timeout_us;
 
   for ( ; ; ) {
     FD_ZERO(&fdset);
     FD_SET(s, &fdset);
     
-    tmv.tv_sec  = timeout_us / 1000000;
-    tmv.tv_usec = timeout_us % 1000000;
-    
     do {
+      tmv.tv_sec  = timeout_left / 1000000;
+      tmv.tv_usec = timeout_left % 1000000;
+    
       rv = select(s+1, &fdset, NULL, NULL, &tmv);
-    } while ( rv == -1 && errno == EINTR );
+      err = errno;
+
+      gettimeofday(&t1, NULL);
+
+      dt = (t1.tv_sec - t0.tv_sec)*1000000 + (t1.tv_usec - t0.tv_usec);
+      *timeout_us_p = timeout_left = ( dt >= timeout_us ) ? 1 : (timeout_us - dt);
+    } while ( rv == -1 && err == EINTR );
     
     if ( rv == 0 ) {
       timer(0);			/* Should not return */
@@ -234,6 +248,7 @@ static int recv_time(int s, void *rbuf, int len, unsigned int flags,
     rv = recv(s, rbuf, len, flags);
     err = errno;
     set_socket_nonblock(s, 0);
+    
     if ( rv < 0 ) {
       if ( E_WOULD_BLOCK(err) || err == EINTR ) {
 	continue;		/* Once again, with feeling... */
@@ -270,6 +285,7 @@ main(int argc, char **argv)
 #ifdef WITH_REGEX
   char *rewrite_file = NULL;
 #endif
+  u_short tp_opcode;
 
   /* basename() is way too much of a pain from a portability standpoint */
 
@@ -278,7 +294,7 @@ main(int argc, char **argv)
   
   openlog(__progname, LOG_PID|LOG_NDELAY, LOG_DAEMON);
   
-  while ((c = getopt(argc, argv, "cspvVla:u:U:r:t:T:m:")) != -1)
+  while ((c = getopt(argc, argv, "cspvVla:B:u:U:r:t:T:m:")) != -1)
     switch (c) {
     case 'c':
       cancreate = 1;
@@ -298,6 +314,17 @@ main(int argc, char **argv)
     case 't':
       waittime = atoi(optarg);
       break;
+    case 'B':
+      {
+	char *vp;
+	max_blksize = (unsigned int)strtoul(optarg, &vp, 10);
+	if ( max_blksize < 512 || max_blksize > MAX_SEGSIZE || *vp ) {
+	  syslog(LOG_ERR, "Bad maximum blocksize value (range 512-%d): %s",
+		 MAX_SEGSIZE, optarg);
+	  exit(EX_USAGE);
+	}
+      }
+      break;
     case 'T':
       {
 	char *vp;
@@ -458,7 +485,7 @@ main(int argc, char **argv)
 	exit(EX_OSERR);
       }
       nfd = open("/dev/null", O_RDWR);
-      if ( nfd >= 0 ) {
+      if ( nfd >= 3 ) {
 #ifdef HAVE_DUP2
 	dup2(nfd, 0);
 	dup2(nfd, 1);
@@ -469,13 +496,11 @@ main(int argc, char **argv)
 	close(2); dup(nfd);
 #endif
 	close(nfd);
-      } else {
+      } else if ( nfd < 0 ) {
 	close(0); close(1); close(2);
       }
 #ifdef HAVE_SETSID
-#ifndef __CYGWIN__		/* Kills the process on Cygwin? */
       setsid();
-#endif
 #endif
     }
   } else {
@@ -677,26 +702,26 @@ main(int argc, char **argv)
     exit(EX_IOERR);
   }
   tp = (struct tftphdr *)buf;
-  tp->th_opcode = ntohs(tp->th_opcode);
-  if (tp->th_opcode == RRQ || tp->th_opcode == WRQ)
+  tp_opcode = ntohs(tp->th_opcode);
+  if (tp_opcode == RRQ || tp_opcode == WRQ)
     tftp(tp, n);
   exit(0);
 }
 
 char   *rewrite_access(char *, int, const char **);
-int	validate_access(char *, int, struct formats *);
+int	validate_access(char *, int, struct formats *, const char **);
 void	tftp_sendfile(struct formats *, struct tftphdr *, int);
 void	tftp_recvfile(struct formats *, struct tftphdr *, int);
 
 struct formats {
   const char *f_mode;
   char	*(*f_rewrite)(char *, int, const char **);
-  int	(*f_validate)(char *, int, struct formats *);
+  int	(*f_validate)(char *, int, struct formats *, const char **);
   void	(*f_send)(struct formats *, struct tftphdr *, int);
   void	(*f_recv)(struct formats *, struct tftphdr *, int);
   int	f_convert;
 } formats[] = {
-  { "netascii",   rewrite_access, validate_access, tftp_sendfile, tftp_recvfile, 1 },
+  { "netascii", rewrite_access, validate_access, tftp_sendfile, tftp_recvfile, 1 },
   { "octet",	rewrite_access, validate_access, tftp_sendfile, tftp_recvfile, 0 },
   { NULL, NULL, NULL, NULL, NULL, 0 }
 };
@@ -707,25 +732,28 @@ struct formats {
 int
 tftp(struct tftphdr *tp, int size)
 {
-  char *cp;
+  char *cp, *end;
   int argn, ecode;
   struct formats *pf = NULL;
   char *origfilename;
   char *filename, *mode = NULL;
-  const char *maperrmsg;
+  const char *errmsgptr;
+  u_short tp_opcode = ntohs(tp->th_opcode);
   
   char *val = NULL, *opt = NULL;
   char *ap = ackbuf + 2;
-  
-  ((struct tftphdr *)ackbuf)->th_opcode = ntohs(OACK);
+
+  ((struct tftphdr *)ackbuf)->th_opcode = htons(OACK);
   
   origfilename = cp = (char *) &(tp->th_stuff);
   argn = 0;
   
-  while ( cp < buf + size && *cp ) {
+  end = (char *)tp + size;
+
+  while ( cp < end && *cp ) {
     do {
       cp++;
-    } while (cp < buf + size && *cp);
+    } while (cp < end && *cp);
     
     if ( *cp ) {
       nak(EBADOP, "Request not null-terminated");
@@ -747,23 +775,23 @@ tftp(struct tftphdr *tp, int size)
 	exit(0);
       }
       if ( !(filename =
-	     (*pf->f_rewrite)(origfilename, tp->th_opcode, &maperrmsg)) ) {
-	nak(EACCESS, maperrmsg); /* File denied by mapping rule */
+	     (*pf->f_rewrite)(origfilename, tp_opcode, &errmsgptr)) ) {
+	nak(EACCESS, errmsgptr); /* File denied by mapping rule */
 	exit(0);
       }
       if ( verbosity >= 1 ) {
 	if ( filename == origfilename || !strcmp(filename, origfilename) )
 	  syslog(LOG_NOTICE, "%s from %s filename %s\n",
-		 tp->th_opcode == WRQ ? "WRQ" : "RRQ",
+		 tp_opcode == WRQ ? "WRQ" : "RRQ",
 		 inet_ntoa(from.sin_addr), filename);
 	else
 	  syslog(LOG_NOTICE, "%s from %s filename %s remapped to %s\n",
-		 tp->th_opcode == WRQ ? "WRQ" : "RRQ",
+		 tp_opcode == WRQ ? "WRQ" : "RRQ",
 		 inet_ntoa(from.sin_addr), origfilename, filename);
       }		   
-      ecode = (*pf->f_validate)(filename, tp->th_opcode, pf);
+      ecode = (*pf->f_validate)(filename, tp_opcode, pf, &errmsgptr);
       if (ecode) {
-	nak(ecode, NULL);
+	nak(ecode, errmsgptr);
 	exit(0);
       }
       opt = ++cp;
@@ -781,12 +809,12 @@ tftp(struct tftphdr *tp, int size)
   }
   
   if ( ap != (ackbuf+2) ) {
-    if ( tp->th_opcode == WRQ )
+    if ( tp_opcode == WRQ )
       (*pf->f_recv)(pf, (struct tftphdr *)ackbuf, ap-ackbuf);
     else
       (*pf->f_send)(pf, (struct tftphdr *)ackbuf, ap-ackbuf);
   } else {
-    if (tp->th_opcode == WRQ)
+    if (tp_opcode == WRQ)
       (*pf->f_recv)(pf, NULL, 0);
     else
       (*pf->f_send)(pf, NULL, 0);
@@ -813,8 +841,8 @@ set_blksize(char *val, char **ret)
   
   if (sz < 8)
     return(0);
-  else if (sz > MAX_SEGSIZE)
-    sz = MAX_SEGSIZE;
+  else if (sz > max_blksize)
+    sz = max_blksize;
   
   segsize = sz;
   sprintf(*ret = b_ret, "%u", sz);
@@ -841,8 +869,8 @@ set_blksize2(char *val, char **ret)
   
   if (sz < 8)
     return(0);
-  else if (sz > MAX_SEGSIZE)
-    sz = MAX_SEGSIZE;
+  else if (sz > max_blksize)
+    sz = max_blksize;
   
   /* Convert to a power of two */
   if ( sz & (sz-1) ) {
@@ -1034,7 +1062,8 @@ FILE *file;
  * given as we have no login directory.
  */
 int
-validate_access(char *filename, int mode, struct formats *pf)
+validate_access(char *filename, int mode,
+		struct formats *pf, const char **errmsg)
 {
   struct stat stbuf;
   int   i, len;
@@ -1044,10 +1073,14 @@ validate_access(char *filename, int mode, struct formats *pf)
   char stdio_mode[3];
   
   tsize_ok = 0;
+  *errmsg = NULL;
   
   if (!secure) {
-    if (*filename != '/')
+    if (*filename != '/') {
+      *errmsg = "Only absolute filenames allowed";
       return (EACCESS);
+    }
+
     /*
      * prevent tricksters from getting around the directory
      * restrictions
@@ -1055,15 +1088,19 @@ validate_access(char *filename, int mode, struct formats *pf)
     len = strlen(filename);
     for ( i = 1 ; i < len-3 ; i++ ) {
       cp = filename + i;
-      if ( *cp == '.' && memcmp(cp-1, "/../", 4) == 0)
+      if ( *cp == '.' && memcmp(cp-1, "/../", 4) == 0 ) {
+	*errmsg = "Reverse path not allowed";
 	return(EACCESS);
+      }
     }
 
     for (dirp = dirs; *dirp; dirp++)
       if (strncmp(filename, *dirp, strlen(*dirp)) == 0)
 	break;
-    if (*dirp==0 && dirp!=dirs)
+    if (*dirp==0 && dirp!=dirs) {
+      *errmsg = "Forbidden directory";
       return (EACCESS);
+    }
   }
   
   /*
@@ -1088,7 +1125,7 @@ validate_access(char *filename, int mode, struct formats *pf)
     case EEXIST:
       return EEXISTS;
     default:
-      return EACCESS;
+      return errno+100;
     }
   }
 
@@ -1096,20 +1133,26 @@ validate_access(char *filename, int mode, struct formats *pf)
     exit(EX_OSERR);		/* This shouldn't happen */
 
   if (mode == RRQ) {
-    if ( !unixperms && (stbuf.st_mode & (S_IREAD >> 6)) == 0 )
+    if ( !unixperms && (stbuf.st_mode & (S_IREAD >> 6)) == 0 ) {
+      *errmsg = "File must have global read permissions";
       return (EACCESS);
+    }
     tsize = stbuf.st_size;
     /* We don't know the tsize if conversion is needed */
     tsize_ok = !pf->f_convert;
   } else {
     if ( !unixperms ) {
-      if ( (stbuf.st_mode & (S_IWRITE >> 6)) == 0 )
+      if ( (stbuf.st_mode & (S_IWRITE >> 6)) == 0 ) {
+	*errmsg = "File must have global write permissions";
 	return (EACCESS);
+      }
 
       /* We didn't get to truncate the file at open() time */
 #ifdef HAVE_FTRUNCATE
-      if ( ftruncate(fd, (off_t)0) )
+      if ( ftruncate(fd, (off_t)0) ) {
+	*errmsg = "Cannot reset file size";
 	return(EACCESS);
+      }
 #endif
     }
     tsize = 0;
@@ -1136,10 +1179,9 @@ tftp_sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   struct tftphdr *dp;
   struct tftphdr *ap;		/* ack packet */
   static u_short block = 1;	/* Static to avoid longjmp funnies */
+  u_short ap_opcode, ap_block;
   int size, n;
   
-  ap = (struct tftphdr *)ackbuf;
-  
   if (oap) {
     timeout = rexmtval;
     (void)sigsetjmp(timeoutbuf,1);
@@ -1149,20 +1191,21 @@ tftp_sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
       goto abort;
     }
     for ( ; ; ) {
-      n = recv_time(peer, ackbuf, sizeof(ackbuf), 0, timeout);
+      n = recv_time(peer, ackbuf, sizeof(ackbuf), 0, &timeout);
       if (n < 0) {
 	syslog(LOG_WARNING, "tftpd: read: %m\n");
 	goto abort;
       }
-      ap->th_opcode = ntohs((u_short)ap->th_opcode);
-      ap->th_block = ntohs((u_short)ap->th_block);
+      ap = (struct tftphdr *)ackbuf;
+      ap_opcode = ntohs((u_short)ap->th_opcode);
+      ap_block = ntohs((u_short)ap->th_block);
       
-      if (ap->th_opcode == ERROR) {
+      if (ap_opcode == ERROR) {
 	syslog(LOG_WARNING, "tftp: client does not accept options\n");
 	goto abort;
       }
-      if (ap->th_opcode == ACK) {
-	if (ap->th_block == 0)
+      if (ap_opcode == ACK) {
+	if (ap_block == 0)
 	  break;
 	/* Resynchronize with the other side */
 	(void)synchnet(peer);
@@ -1189,19 +1232,20 @@ tftp_sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
     }
     read_ahead(file, pf->f_convert);
     for ( ; ; ) {
-      n = recv_time(peer, ackbuf, sizeof (ackbuf), 0, timeout);
+      n = recv_time(peer, ackbuf, sizeof (ackbuf), 0, &timeout);
       if (n < 0) {
 	syslog(LOG_WARNING, "tftpd: read(ack): %m");
 	goto abort;
       }
-      ap->th_opcode = ntohs((u_short)ap->th_opcode);
-      ap->th_block = ntohs((u_short)ap->th_block);
+      ap = (struct tftphdr *)ackbuf;
+      ap_opcode = ntohs((u_short)ap->th_opcode);
+      ap_block = ntohs((u_short)ap->th_block);
       
-      if (ap->th_opcode == ERROR)
+      if (ap_opcode == ERROR)
 	goto abort;
       
-      if (ap->th_opcode == ACK) {
-	if (ap->th_block == block) {
+      if (ap_opcode == ACK) {
+	if (ap_block == block) {
 	  break;
 	}
 				/* Re-synchronize with the other side */
@@ -1241,6 +1285,7 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   static struct tftphdr *ap;    /* ack buffer */
   static u_short block = 0;
   static int acksize;
+  u_short dp_opcode, dp_block;
 
   dp = w_init();
   do {
@@ -1264,22 +1309,22 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
     }
     write_behind(file, pf->f_convert);
     for ( ; ; ) {
-      n = recv_time(peer, dp, PKTSIZE, 0, timeout);
+      n = recv_time(peer, dp, PKTSIZE, 0, &timeout);
       if (n < 0) {		/* really? */
 	syslog(LOG_WARNING, "tftpd: read: %m");
 	goto abort;
       }
-      dp->th_opcode = ntohs((u_short)dp->th_opcode);
-      dp->th_block = ntohs((u_short)dp->th_block);
-      if (dp->th_opcode == ERROR)
+      dp_opcode = ntohs((u_short)dp->th_opcode);
+      dp_block = ntohs((u_short)dp->th_block);
+      if (dp_opcode == ERROR)
 	goto abort;
-      if (dp->th_opcode == DATA) {
-	if (dp->th_block == block) {
+      if (dp_opcode == DATA) {
+	if (dp_block == block) {
 	  break;   /* normal */
 	}
 				/* Re-synchronize with the other side */
 	(void) synchnet(peer);
-	if (dp->th_block == (block-1))
+	if (dp_block == (block-1))
 	  goto send_ack;		/* rexmit */
       }
     }
@@ -1299,12 +1344,12 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   (void) send(peer, ackbuf, 4, 0);
   
   timeout_quit = 1;		/* just quit on timeout */
-  n = recv_time(peer, buf, sizeof (buf), 0, timeout); /* normally times out and quits */
+  n = recv_time(peer, buf, sizeof (buf), 0, &timeout); /* normally times out and quits */
   timeout_quit = 0;
 
   if (n >= 4 &&			/* if read some data */
-      dp->th_opcode == DATA &&    /* and got a data block */
-      block == dp->th_block) {	/* then my last ack was lost */
+      dp_opcode == DATA &&	/* and got a data block */
+      block == dp_block) {	/* then my last ack was lost */
     (void) send(peer, ackbuf, 4, 0);     /* resend final ack */
   }
  abort:
@@ -1339,18 +1384,22 @@ nak(int error, const char *msg)
   
   tp = (struct tftphdr *)buf;
   tp->th_opcode = htons((u_short)ERROR);
-  tp->th_code = htons((u_short)error);
-  if ( !msg ) {
-    if ( error >= 100 ) {
-      /* This is a Unix errno+100 */
+
+  if ( error >= 100 ) {
+    /* This is a Unix errno+100 */
+    if ( !msg )
       msg = strerror(error - 100);
-      tp->th_code = EUNDEF;
-    } else {
-      if ( (unsigned)error >= ERR_CNT )
-	tp->th_code = error = EUNDEF;
+    error = EUNDEF;
+  } else {
+    if ( (unsigned)error >= ERR_CNT )
+      error = EUNDEF;
+
+    if ( !msg )
       msg = errmsgs[error];
-    }
   }
+
+  tp->th_code = htons((u_short)error);
+
   length = strlen(msg)+1;
   memcpy(tp->th_msg, msg, length);
   length += 4;			/* Add space for header */
